import unittest

from hgijson.serialization import PropertyMapping


class TestPropertyMapping(unittest.TestCase):
    """
    Tests for `PropertyMapping`.
    """
    def test_init_with_constructor_name_and_object_property_setter(self):
        self.assertRaises(ValueError, PropertyMapping, object_constructor_parameter_name="a",
                          serialized_property_getter=lambda: None, object_property_setter=lambda: None)

    def test_init_with_constructor_name_but_no_serialized_property_getter_or_name(self):
        self.assertRaises(ValueError, PropertyMapping, object_constructor_parameter_name="a")

    def test_init_with_constructor_argument_modifier_but_no_constructor_parameter_name(self):
        self.assertRaises(ValueError, PropertyMapping, object_constructor_argument_modifier=lambda: None)

    def test_init_with_no_arguments(self):
        property_mapping = PropertyMapping()
        self.assertIsNone(property_mapping.object_property_getter)
        self.assertIsNone(property_mapping.object_property_setter)
        self.assertIsNone(property_mapping.serialized_property_getter)
        self.assertIsNone(property_mapping.serialized_property_setter)
        self.assertIsNone(property_mapping.object_constructor_parameter_name)

    def test_init_with_object_and_serialized_getters_and_setters(self):
        property_mapping = PropertyMapping(
                serialized_property_getter=lambda: None, serialized_property_setter=lambda: None,
                object_property_getter=lambda: None, object_property_setter=lambda: None)
        self.assertIsNotNone(property_mapping.object_property_getter)
        self.assertIsNotNone(property_mapping.object_property_setter)
        self.assertIsNotNone(property_mapping.serialized_property_getter)
        self.assertIsNotNone(property_mapping.serialized_property_setter)
        self.assertIsNone(property_mapping.object_constructor_parameter_name)

    def test_str(self):
        property_mapping = PropertyMapping(object_constructor_parameter_name="a",
                                           object_property_getter=lambda obj: obj.constructor_a,
                                           serialized_property_getter=lambda: None)
        self.assertIsInstance(str(property_mapping), str)

    def test_repr(self):
        property_mapping = PropertyMapping(object_constructor_parameter_name="a",
                                           object_property_getter=lambda obj: obj.constructor_a,
                                           serialized_property_getter=lambda: None)
        self.assertIsInstance(repr(property_mapping), str)


if __name__ == "__main__":
    unittest.main()



import json
import unittest

from hgijson.json_converters._serialization import MappingJSONDecoder, MappingJSONEncoder
from hgijson.json_converters.builders import MappingJSONEncoderClassBuilder, MappingJSONDecoderClassBuilder
from hgijson.json_converters.models import JsonPropertyMapping
from hgijson.tests._models import SimpleModel, ComplexModel, BaseModel
from hgijson.tests.json_converters._helpers import create_complex_model_with_json_representation, \
    create_simple_model_with_json_representation
from hgijson.tests.json_converters._serializers import get_simple_model_json_property_mappings, \
    get_complex_model_json_property_mappings


class _Named(BaseModel):
    def __init__(self):
        super().__init__()
        self.name = None    # type: str


class _Office(_Named):
    def __init__(self):
        super().__init__()


class _Identifiable(BaseModel):
    def __init__(self):
        super().__init__()
        self.id = None    # type: int


class _Employee(_Named, _Identifiable):
    def __init__(self):
        super().__init__()
        self.title = None
        self.office = None


class _Container(BaseModel):
    def __init__(self, colour: str, container: "_Container" = None):
        self.colour = colour
        self.contains = container


class TestMappingJSONEncoderClassBuilder(unittest.TestCase):
    """
    Tests for `MappingJSONEncoderClassBuilder`.
    """
    def setUp(self):
        self.simple_model, self.simple_model_as_json = create_simple_model_with_json_representation()
        self.complex_model, self.complex_model_as_json = create_complex_model_with_json_representation()

    def test_build(self):
        encoder_builder = MappingJSONEncoderClassBuilder()
        encoder_builder.target_cls = SimpleModel
        encoder_builder.mappings = get_simple_model_json_property_mappings()

        encoder_cls = encoder_builder.build()
        encoder = encoder_cls()     # type: MappingJSONEncoder

        encoded = encoder.default(self.simple_model)
        self.assertDictEqual(encoded, self.simple_model_as_json)

    def test_build_with_superclass(self):
        SimpleModelJSONEncoder = MappingJSONEncoderClassBuilder(
            SimpleModel, get_simple_model_json_property_mappings()).build()

        encoder_builder = MappingJSONEncoderClassBuilder()
        encoder_builder.target_cls = ComplexModel
        encoder_builder.mappings = get_complex_model_json_property_mappings()
        encoder_builder.superclasses = (SimpleModelJSONEncoder, )

        encoder_cls = encoder_builder.build()
        encoder = encoder_cls()     # type: MappingJSONEncoder

        encoded = encoder.default(self.complex_model)
        self.assertDictEqual(encoded, self.complex_model_as_json)

    def test_build_with_superclass_as_string(self):
        ContainerJSONEncoder = MappingJSONEncoderClassBuilder(_Container, [
            JsonPropertyMapping("colour", "colour", object_constructor_parameter_name="colour"),
            JsonPropertyMapping("contains", "contains", encoder_cls=lambda: ContainerJSONEncoder, optional=True)
        ]).build()
        encoder = ContainerJSONEncoder()
        container = _Container("red", _Container("blue"))
        self.assertEqual({"colour": "red", "contains": {"colour": "blue"}}, encoder.default(container))

    def test_build_with_multiple_superclasses(self):
        NamedJSONEncoder = MappingJSONEncoderClassBuilder(_Named, [
            JsonPropertyMapping("name", "name")
        ]).build()
        OfficeJSONEncoder = MappingJSONEncoderClassBuilder(_Office, [], (NamedJSONEncoder, )).build()
        IdentifiableJSONEncoder = MappingJSONEncoderClassBuilder(_Identifiable, [
            JsonPropertyMapping("id", "id")
        ]).build()
        EmployeeJSONEncoder = MappingJSONEncoderClassBuilder(_Employee, [
            JsonPropertyMapping("title", "title"),
            JsonPropertyMapping("office", "office", encoder_cls=OfficeJSONEncoder),
        ], (NamedJSONEncoder, IdentifiableJSONEncoder)).build()

        office = _Office()
        office.name = "Cambridge"

        employee = _Employee()
        employee.name = "Bob"
        employee.id = 42
        employee.title = "Software Dev"
        employee.office = office

        employee_as_json = {
            "name": employee.name,
            "id": employee.id,
            "title": employee.title,
            "office": {
                "name": office.name
            }
        }
        self.assertEqual(EmployeeJSONEncoder().default(employee), employee_as_json)

    def test_build_with_none_property(self):
        NamedJSONEncoder = MappingJSONEncoderClassBuilder(_Named, [
            JsonPropertyMapping("name", "name")
        ]).build()
        OfficeJSONEncoder = MappingJSONEncoderClassBuilder(_Office, [], (NamedJSONEncoder,)).build()
        EmployeeJSONEncoder = MappingJSONEncoderClassBuilder(_Employee, [
            JsonPropertyMapping("office", "office", encoder_cls=OfficeJSONEncoder),
        ], (NamedJSONEncoder, )).build()

        employee = _Employee()
        employee.name = "Bob"
        employee.office = None

        employee_as_json = {
            "name": employee.name,
            "office": None
        }
        self.assertEqual(employee_as_json, EmployeeJSONEncoder().default(employee))


class TestMappingJSONDecoderClassBuilder(unittest.TestCase):
    """
    Tests for `MappingJSONDecoderClassBuilder`.
    """
    def setUp(self):
        self.simple_model, self.simple_model_as_json = create_simple_model_with_json_representation()
        self.complex_model, self.complex_model_as_json = create_complex_model_with_json_representation()

    def test_build(self):
        decoder_builder = MappingJSONDecoderClassBuilder()
        decoder_builder.target_cls = SimpleModel
        decoder_builder.mappings = get_simple_model_json_property_mappings()

        decoder_cls = decoder_builder.build()
        decoder = decoder_cls()     # type: MappingJSONDecoder

        decoded = decoder.decode(json.dumps(self.simple_model_as_json))
        self.assertEqual(decoded, self.simple_model)

    def test_build_with_superclass(self):
        SimpleModelJSONDecoder = MappingJSONDecoderClassBuilder(
            target_cls=SimpleModel, mappings=get_simple_model_json_property_mappings()).build()

        decoder_builder = MappingJSONDecoderClassBuilder()
        decoder_builder.superclasses = (SimpleModelJSONDecoder, )
        decoder_builder.target_cls = ComplexModel
        decoder_builder.mappings = get_complex_model_json_property_mappings()

        decoder_cls = decoder_builder.build()
        decoder = decoder_cls()     # type: MappingJSONDecoder

        decoded = decoder.decode(json.dumps(self.complex_model_as_json))
        self.assertEqual(decoded, self.complex_model)

    def test_build_with_superclass_string(self):
        ContainerJSONDecoder = MappingJSONDecoderClassBuilder(_Container, [
            JsonPropertyMapping("colour", "colour", object_constructor_parameter_name="colour"),
            JsonPropertyMapping("contains", "contains", decoder_cls=lambda: ContainerJSONDecoder, optional=True)
        ]).build()
        decoder = ContainerJSONDecoder()
        container = _Container("red", _Container("blue"))
        self.assertEqual(container, decoder.decode(json.dumps({"colour": "red", "contains": {"colour": "blue"}})))

    def test_build_with_multiple_superclasses(self):
        NamedJSONDecoder = MappingJSONDecoderClassBuilder(_Named, [
            JsonPropertyMapping("name", "name")
        ]).build()
        OfficeJSONDecoder = MappingJSONDecoderClassBuilder(_Office, [], (NamedJSONDecoder,)).build()
        IdentifiableJSONDecoder = MappingJSONDecoderClassBuilder(_Identifiable, [
            JsonPropertyMapping("id", "id")
        ]).build()
        EmployeeJSONDecoder = MappingJSONDecoderClassBuilder(_Employee, [
            JsonPropertyMapping("title", "title"),
            JsonPropertyMapping("office", "office", decoder_cls=OfficeJSONDecoder)
        ], (NamedJSONDecoder, IdentifiableJSONDecoder)).build()

        office = _Office()
        office.name = "Cambridge"

        employee = _Employee()
        employee.name = "Bob"
        employee.id = 42
        employee.title = "Software Dev"
        employee.office = office

        employee_as_json = {
            "name": employee.name,
            "id": employee.id,
            "title": employee.title,
            "office": {
                "name": office.name
            }
        }
        employee_as_json_string = json.dumps([employee_as_json])
        self.assertEqual(EmployeeJSONDecoder().decode(employee_as_json_string), [employee])

    def test_build_with_none_property(self):
        NamedJSONDecoder = MappingJSONDecoderClassBuilder(_Named, [
            JsonPropertyMapping("name", "name")
        ]).build()
        OfficeJSONDecoder = MappingJSONDecoderClassBuilder(_Office, [], (NamedJSONDecoder,)).build()
        EmployeeJSONDecoder = MappingJSONDecoderClassBuilder(_Employee, [
            JsonPropertyMapping("office", "office", decoder_cls=OfficeJSONDecoder)
        ], (NamedJSONDecoder, )).build()

        employee = _Employee()
        employee.name = "Bob"
        employee.office = None

        employee_as_json = {
            "name": employee.name,
            "id": employee.id,
            "title": employee.title,
            "office": None
        }
        employee_as_json_string = json.dumps([employee_as_json])
        self.assertEqual(EmployeeJSONDecoder().decode(employee_as_json_string), [employee])


if __name__ == "__main__":
    unittest.main()



from typing import TypeVar, Dict, List, Tuple, Union

SerializableType = TypeVar("Serializable")

# TODO: Why is this needed?
PrimitiveUnionType = TypeVar("PrimitiveUnion")

PrimitiveJsonType = Union[Dict, List, Tuple, str, int, float, bool, None]



from json import JSONDecoder, JSONEncoder
from typing import Callable, Any, Dict, Union, Iterable

from hgijson.json_converters._converters import json_decoder_to_deserializer, json_encoder_to_serializer
from hgijson.serialization import PropertyMapping


class JsonPropertyMapping(PropertyMapping):
    """
    Model of a mapping between a json property and a property of an object.
    """
    @property
    def json_property_getter(self) -> Callable[[Dict], Any]:
        return self.serialized_property_getter

    @json_property_getter.setter
    def json_property_getter(self, getter: Callable[[Dict], Any]):
        self.serialized_property_getter = getter

    @property
    def json_property_setter(self) -> Callable[[Any, Any], None]:
        return self.serialized_property_setter

    @json_property_setter.setter
    def json_property_setter(self, setter: Callable[[Any, Any], None]):
        self.serialized_property_setter = setter

    def __init__(
            self, json_property_name=None, object_property_name: str=None, object_constructor_parameter_name: str=None,
            *, object_constructor_argument_modifier: Callable[[Any], Any]=None,
            json_property_getter: Callable[[Dict], Any]=None, json_property_setter: Callable[[Any, Any], None]=None,
            object_property_getter: Callable[[Any], Any]=None, object_property_setter: Callable[[Any, Any], None]=None,
            encoder_cls: Union[type, Callable[[], type]]=JSONEncoder,
            decoder_cls: Union[type, Callable[[], type]]=JSONDecoder, 
            optional: bool=False,
            collection_factory: Callable[[Iterable], Any]=lambda items: list(items),
            collection_iter: Callable[[Any], Iterable]=lambda collection: iter(collection)):
        """
        Constructor.
        :param json_property_name:
        :param object_property_name:
        :param object_constructor_parameter_name:
        :param object_constructor_argument_modifier:
        :param json_property_getter:
        :param json_property_setter:
        :param object_property_getter:
        :param object_property_setter:
        :param encoder_cls:
        :param decoder_cls:
        :param optional:
        """
        if json_property_name is not None:
            if json_property_getter is not None and json_property_setter is not None:
                raise ValueError("Redundant `json_property_name` argument given. It has been specified that a "
                                 "serialized property is to be used via the given property name and both a setter and "
                                 "getter of this property has been provided. To avoid confusion, the serialized "
                                 "property cannot be specified in this case.")

            if json_property_getter is None:
                def json_property_getter(obj_as_json: dict):
                    if json_property_name not in obj_as_json:
                        if optional:
                            return None
                        else:
                            raise KeyError("No value associated to the non-optional key `%s` in the input JSON: `%s`"
                                             % (json_property_name, obj_as_json))
                    return obj_as_json[json_property_name]

            if json_property_setter is None:
                def json_property_setter(obj_as_json: dict, value: Any):
                    obj_as_json[json_property_name] = value

        if object_property_name is not None:
            if object_property_getter is not None and object_property_setter is not None:
                raise ValueError("Redundant `object_property_name` argument given. It has been specified that an "
                                 "object property is to be used via the given property name and both a setter and "
                                 "getter of this object property has been provided. To avoid confusion, the object "
                                 "property cannot be specified in this case.")

            if object_property_getter is None:
                def object_property_getter(obj: Any) -> Any:
                    return getattr(obj, object_property_name)

            if object_property_setter is None and object_constructor_parameter_name is None:
                def object_property_setter(obj: Any, value: Any):
                    if not hasattr(obj, object_property_name):
                        raise AttributeError("Object `%s` does not have the attribute `%s`"
                                             % (obj, object_property_name))
                    setattr(obj, object_property_name, value)

        encoder_as_serializer_cls = json_encoder_to_serializer(encoder_cls)
        decoder_as_serializer_cls = json_decoder_to_deserializer(decoder_cls)

        super().__init__(serialized_property_getter=json_property_getter,
                         serialized_property_setter=json_property_setter,
                         object_property_getter=object_property_getter,
                         object_property_setter=object_property_setter,
                         object_constructor_parameter_name=object_constructor_parameter_name,
                         object_constructor_argument_modifier=object_constructor_argument_modifier,
                         serializer_cls=encoder_as_serializer_cls, deserializer_cls=decoder_as_serializer_cls,
                         optional=optional,
                         collection_factory=collection_factory, collection_iter=collection_iter)



import json
from json import JSONEncoder, JSONDecoder
from typing import Iterable

from hgijson.custom_types import PrimitiveJsonType
from hgijson.json_converters._serialization import MappingJSONDecoder, MappingJSONEncoder
from hgijson.json_converters.models import JsonPropertyMapping
from hgijson.tests._models import SimpleModel, ComplexModel


def get_simple_model_json_property_mappings() -> Iterable[JsonPropertyMapping]:
    return [
        JsonPropertyMapping("serialized_a", "a"),
        JsonPropertyMapping("serialized_b", "b", "constructor_b")
    ]


def get_complex_model_json_property_mappings() -> Iterable[JsonPropertyMapping]:
    return [
        JsonPropertyMapping("serialized_b", "b", "constructor_b"),
        JsonPropertyMapping("serialized_c", "c"),
        JsonPropertyMapping("serialized_d", "d",
                            encoder_cls=SimpleModelMappingJSONEncoder, decoder_cls=SimpleModelMappingJSONDecoder),
        JsonPropertyMapping("serialized_e", "e"),
        JsonPropertyMapping("serialized_f", "f"),
        JsonPropertyMapping("serialized_g", "g"),
        JsonPropertyMapping("serialized_h", "h"),
        JsonPropertyMapping("serialized_i", "i", collection_factory=set)
    ]


class BasicSimpleModelJSONEncoder(JSONEncoder):
    """
    Basic JSON encoder for `SimpleModel`.
    """
    def default(self, to_encode: SimpleModel) -> PrimitiveJsonType:
        return {
            "serialized_a": to_encode.a,
            "serialized_b": to_encode.b
        }


class BasicSimpleModelJSONDecoder(JSONDecoder):
    """
    Basic JSON decoder for `SimpleModel`.
    """
    def decode(self, to_decode_as_str: str, *args, **kwargs) -> object:
        to_decode_as_dict = json.loads(to_decode_as_str)

        simple_model = SimpleModel()
        simple_model.a = to_decode_as_dict["serialized_a"]
        simple_model.b = to_decode_as_dict["serialized_b"]
        return simple_model


class SimpleModelMappingJSONEncoder(MappingJSONEncoder):
    """
    Mapping JSON encoder for `SimpleModel`.
    """
    def _get_property_mappings(self) -> Iterable[JsonPropertyMapping]:
        return get_simple_model_json_property_mappings()

    def _get_serializable_cls(self) -> type:
        return SimpleModel


class SimpleModelMappingJSONDecoder(MappingJSONDecoder):
    """
    Mapping JSON decoder for `SimpleModel`.
    """
    def _get_property_mappings(self) -> Iterable[JsonPropertyMapping]:
        return get_simple_model_json_property_mappings()

    def _get_deserializable_cls(self) -> type:
        return SimpleModel


class ComplexModelMappingJSONEncoder(SimpleModelMappingJSONEncoder):
    """
    Mapping JSON encoder for `ComplexModel`.
    """
    def _get_property_mappings(self) -> Iterable[JsonPropertyMapping]:
        # FIXME: Merge correctly, ensure subclass mappings override superclass mappings
        return list(super()._get_property_mappings()) + list(get_complex_model_json_property_mappings())

    def _get_serializable_cls(self) -> type:
        return ComplexModel


class ComplexModelMappingJSONDecoder(SimpleModelMappingJSONDecoder):
    """
    Mapping JSON decoder for `ComplexModel`.
    """
    def _get_property_mappings(self) -> Iterable[JsonPropertyMapping]:
        # FIXME: Merge correctly, ensure subclass mappings override superclass mappings
        return list(super()._get_property_mappings()) + list(get_complex_model_json_property_mappings())

    def _get_deserializable_cls(self) -> type:
        return ComplexModel



from abc import ABCMeta, abstractmethod
from typing import Any, Generic, Union, Dict, List, Optional, Iterable, Type, Callable

from hgijson.custom_types import SerializableType, PrimitiveUnionType, PrimitiveJsonType


class PropertyMapping:
    """
    Model of a mapping between a json property and a property of an object.
    """
    def __init__(
            self, *,
            serialized_property_getter: Callable[[Dict], Any]=None,
            serialized_property_setter: Callable[[Any, Any], None]=None,
            object_property_getter: Callable[[Any], Any]=None, object_property_setter: Callable[[Any, Any], None]=None,
            object_constructor_parameter_name: str = None,
            object_constructor_argument_modifier: Callable[[Any], Any] = None,
            serializer_cls: Type["Serializer"]=None,
            deserializer_cls: Type["Deserializer"]=None,
            optional: bool=False,
            collection_factory: Callable[[Iterable], Any]=lambda items: list(items),
            collection_iter: Callable[[Any], Iterable]=lambda collection: iter(collection)):
        """
        Constructor.
        :param object_property_name: defines the object property to assign the value returned by
        `serialized_property_getter` to using a simple `object_property_getter` function
        :param object_constructor_parameter_name: defines the constructor parameter to bind the value returned by
        `serialized_property_getter` to using a simple `serialized_property_getter` function
        :param object_constructor_argument_modifier: constructor argument modification function
        :param serialized_property_getter: method that, given the serialized object, gets the property of interest to
        this mapping
        :param serialized_property_setter: method that, given the serialized object, sets a serialized property to the
        value returned by `object_property_getter`
        :param object_property_getter: method that, given the object, gets the property of interest to this mapping
        :param object_property_setter: method that, given the object, sets a property to the value returned by
        `serialized_property_getter`
        :param serializer_cls: class to serialize the value returned by `object_property_getter` (defaults to
        PrimitiveSerializer)
        :param deserializer_cls: class to deserialize the value returned by `serialized_property_getter` (defaults to
        PrimitiveDeserializer)
        :param optional: whether the property is optional - will ignore if `None` in serialized representation and will
        not serialize if `None` in object
        """
        if object_constructor_parameter_name is not None:
            if serialized_property_getter is None:
                raise ValueError("`serialized_property_getter` must be defined alongside "
                                 "`object_constructor_parameter_name`.")

            if object_property_setter is not None and object_constructor_parameter_name is not None:
                raise ValueError("`object_property_setter` cannot be defined if `object_constructor_parameter_name` is "
                                 "given as the serialized data will be injected into the object as an argument. If an "
                                 "object property should also be set using a setter, define another mapping. If the "
                                 "constructor argument needs to be modified, use "
                                 "`object_constructor_argument_modifier`.")

            if object_constructor_argument_modifier is None:
                def object_constructor_argument_modifier(argument: Any) -> Any:
                    return argument
        else:
            if object_constructor_argument_modifier is not None:
                raise ValueError("`object_constructor_argument_modifier` cannot be used without "
                                 "`object_constructor_parameter_name` being set.")

        from hgijson.serializers import PrimitiveSerializer, PrimitiveDeserializer

        self.serialized_property_getter = serialized_property_getter
        self.serialized_property_setter = serialized_property_setter
        self.object_constructor_parameter_name = object_constructor_parameter_name
        self.object_constructor_argument_modifier = object_constructor_argument_modifier
        self.object_property_getter = object_property_getter
        self.object_property_setter = object_property_setter
        self.serializer_cls = serializer_cls if serializer_cls is not None else PrimitiveSerializer
        self.deserializer_cls = deserializer_cls if deserializer_cls is not None else PrimitiveDeserializer
        self.optional = optional
        self.collection_factory = collection_factory
        self.collection_iter = collection_iter

    def __str__(self) -> str:
        string_builder = []
        for property, value in vars(self).items():
            string_builder.append("%s: %s" % (property, value))
        string_builder = sorted(string_builder)
        return "{ %s }" % ', '.join(string_builder)

    def __repr__(self) -> str:
        return "<%s object at %s: %s>" % (type(self), id(self), str(self))


class Serializer(Generic[SerializableType, PrimitiveUnionType], metaclass=ABCMeta):
    """
    Serializer that uses a mapping that describes how the process should occur for the type of serializable object that
    this serializer handles.
    """
    @abstractmethod
    def _create_serializer_of_type(self, serializer_type: Type) -> "Serializer":
        """
        Create an instance of an serializer of the given type.
        :param serializer_type: the type of serializer to instantiate (a subclass of `Serializer`)
        :return: the created serializer
        """

    @abstractmethod
    def _create_serialized_container(self) -> Any:
        """
        Create the container in which serialized representation is built in
        :return: the container
        """

    def __init__(self, property_mappings: Iterable[PropertyMapping]):
        """
        Constructor.
        :param property_mappings: the property mappings (of type `List[PropertyMapping]`)
        """
        self._property_mappings = property_mappings     # type: Iterable[PropertyMapping]
        self._serializers_cache = dict()    # type: Dict[type, Serializer]

    def serialize(self, serializable: Optional[Union[SerializableType, List[SerializableType]]]) \
            -> PrimitiveJsonType:
        """
        Serializes the given serializable object or collection of serializable objects.
        :param serializable: the object or objects to serialize
        :return: a serialization of the given object
        """
        if serializable is None:
            # Implements #17
            return None
        elif isinstance(serializable, List):
            return [self.serialize(item) for item in serializable]
        else:
            serialized = self._create_serialized_container()

            for mapping in self._property_mappings:
                if mapping.object_property_getter is not None and mapping.serialized_property_setter is not None:
                    value = mapping.object_property_getter(serializable)
                    if not (mapping.optional and value is None):
                        if isinstance(value, type(mapping.collection_factory([]))):
                            value = list(mapping.collection_iter(value))
                        encoded_value = self._serialize_property_value(value, mapping.serializer_cls)
                        mapping.serialized_property_setter(serialized, encoded_value)

            return serialized

    def _serialize_property_value(self, to_serialize: Any, serializer_cls: Type) -> Any:
        """
        Serializes the given value using the given serializer.
        :param to_serialize: the value to deserialize
        :param serializer_cls: the type of serializer to use
        :return: serialized value
        """
        serializer = self._create_serializer_of_type_with_cache(serializer_cls)
        assert serializer is not None
        return serializer.serialize(to_serialize)

    def _create_serializer_of_type_with_cache(self, serializer_type: Type) -> "Serializer":
        """
        Creates a deserializer of the given type, exploiting a cache.
        :param serializer_type: the type of deserializer to create
        :return: the created serializer
        """
        if serializer_type not in self._serializers_cache:
            self._serializers_cache[serializer_type] = self._create_serializer_of_type(serializer_type)
        return self._serializers_cache[serializer_type]


class Deserializer(Generic[SerializableType, PrimitiveUnionType], metaclass=ABCMeta):
    """
    Deserializer that uses a mapping that describes how the process should occur for the type of deserializable object
    that this deserializer handles.
    """
    @abstractmethod
    def _create_deserializer_of_type(self, deserializer_type: Type) -> "Deserializer":
        """
        Creates a deserializer of the given type.
        :param deserializer_type: the type of deserializer to create
        :return: the created deserializer
        """

    def __init__(self, property_mappings: Iterable[PropertyMapping], deserializable_cls: Type):
        """
        Constructor.
        :param property_mappings: TODO
        :param deserializable_cls: the class that should be built as a result of deserialization
        """
        self._property_mappings = property_mappings     # type: Iterable[PropertyMapping]
        self._deserializable_cls = deserializable_cls
        self._deserializers_cache = dict()    # type: Dict[type, Deserializer]

    def deserialize(self, to_deserialize: PrimitiveJsonType) \
            -> Optional[Union[SerializableType, List[SerializableType]]]:
        """
        Deserializes the given representation of the serialized object.
        :param to_deserialize: the serialized object as a dictionary
        :return: the deserialized object or collection of deserialized objects
        """
        if to_deserialize is None:
            # Implements #17
            return None
        elif isinstance(to_deserialize, List):
            deserialized = []
            for item in to_deserialize:
                item_deserialized = self.deserialize(item)
                deserialized.append(item_deserialized)
            return deserialized
        else:
            mappings_not_set_in_constructor = []    # type: List[PropertyMapping]

            init_kwargs = dict()    # type: Dict[str, Any]
            for mapping in self._property_mappings:
                if mapping.object_constructor_parameter_name is not None:
                    value = mapping.serialized_property_getter(to_deserialize)
                    if not (mapping.optional and value is None):
                        decoded_value = self._deserialize_property_value(value, mapping.deserializer_cls)
                        if isinstance(decoded_value, list):
                            collection = mapping.collection_factory(decoded_value)
                            decoded_value = collection

                        argument = mapping.object_constructor_argument_modifier(decoded_value)
                        init_kwargs[mapping.object_constructor_parameter_name] = argument
                else:
                    mappings_not_set_in_constructor.append(mapping)

            decoded = self._deserializable_cls(**init_kwargs)
            assert type(decoded) == self._deserializable_cls

            for mapping in mappings_not_set_in_constructor:
                assert mapping.object_constructor_parameter_name is None
                if mapping.serialized_property_getter is not None and mapping.object_property_setter is not None:
                    value = mapping.serialized_property_getter(to_deserialize)
                    if not (mapping.optional and value is None):
                        decoded_value = self._deserialize_property_value(value, mapping.deserializer_cls)
                        if isinstance(decoded_value, list):
                            collection = mapping.collection_factory(decoded_value)
                            decoded_value = collection

                        mapping.object_property_setter(decoded, decoded_value)

            return decoded

    def _deserialize_property_value(self, to_deserialize: PrimitiveJsonType, deserializer_cls: Type) -> Any:
        """
        Deserializes the given value using the given deserializer.
        :param to_deserialize: the value to deserialize
        :param deserializer_cls: the type of deserializer to use
        :return: deserialized value
        """
        deserializer = self._create_deserializer_of_type_with_cache(deserializer_cls)
        assert deserializer is not None
        return deserializer.deserialize(to_deserialize)

    def _create_deserializer_of_type_with_cache(self, deserializer_type: Type) -> "Deserializer":
        """
        Creates a deserializer of the given type, exploiting a cache.
        :param deserializer_type: the type of deserializer to create
        :return: the deserializer
        """
        if deserializer_type not in self._deserializers_cache:
            self._deserializers_cache[deserializer_type] = self._create_deserializer_of_type(deserializer_type)
        return self._deserializers_cache[deserializer_type]



from abc import ABCMeta, abstractmethod
from json import JSONEncoder
from typing import Union, List, Optional

from hgijson.json_converters._serializers import JsonObjectSerializer, JsonObjectDeserializer
from hgijson.json_converters.interfaces import ParsedJSONDecoder
from hgijson.serialization import PropertyMapping
from hgijson.custom_types import PrimitiveJsonType, SerializableType


class PropertyMapper(metaclass=ABCMeta):
    """
    Model of a mapping from a property of a JSON model to a property of a native Python object.
    """
    def _get_property_mappings(self) -> List[PropertyMapping]:
        """
        Gets the property mappings that are to be used in this encoder.
        :return: the property mappings to use in the order in which they should be applied
        """
        return []


class MappingJSONEncoder(JSONEncoder, PropertyMapper, metaclass=ABCMeta):
    """
    JSON encoder that serialises an object based on a mapping of its properties to JSON properties.

    As `json.dumps` requires a type rather than an instance and there is no control given over the instatiation, the
    encoded class and the mappings between the object properties and the json properties cannot be passed through the
    constructor. Instead this class must be subclassed and the subclass must define the relevant constants.
    """
    @abstractmethod
    def _get_serializable_cls(self) -> type:
        """
        Gets the type of class that this encoder will serialize.
        :return: the class the encoder will serialize
        """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._args = args
        self._kwargs = kwargs
        self._serializer_cache = None

    def default(self, serializable: Optional[Union[SerializableType, List[SerializableType]]]) \
            -> PrimitiveJsonType:
        serializer = self._create_serializer()
        if serializable is not None \
                and not isinstance(serializable, List) \
                and not isinstance(serializable, self._get_serializable_cls()):
            return super().default(serializable)

        return serializer.serialize(serializable)

    def _create_serializer(self) -> JsonObjectSerializer:
        """
        Create serializer that is to be used by this encoder
        :return: the serializer
        """
        if self._serializer_cache is None:
            serializer_cls = type(
                "%sInternalSerializer" % type(self),
                (JsonObjectSerializer,),
                {
                    "_JSON_ENCODER_ARGS": self._args,
                    "_JSON_ENCODER_KWARGS": self._kwargs
                }
            )
            self._serializer_cache = serializer_cls(self._get_property_mappings())
        return self._serializer_cache


class MappingJSONDecoder(ParsedJSONDecoder, PropertyMapper, metaclass=ABCMeta):
    """
    JSON decoder that creates an object from JSON based on a mapping from the JSON properties to the object properties,
    mindful that some properties may have to be passed through the constructor.

    Can decode collections of the deserializable object.

    As `json.dumps` requires a type rather than an instance and there is no control given over the instatiation, the
    decoded class and the mappings between the object properties and the json properties cannot be passed through the
    constructor. Instead this class must be subclassed and the subclass must define the relevant constants.
    """
    @abstractmethod
    def _get_deserializable_cls(self) -> type:
        """
        Gets the type of class that this decoder will deserialize.
        :return: the class the decoder will deserialize
        """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._args = args
        self._kwargs = kwargs
        self._deserializer_cache = None

    def decode(self, json_as_string: str, **kwargs) -> SerializableType:
        json_as_dict = super().decode(json_as_string)
        return self.decode_parsed(json_as_dict)

    def decode_parsed(self, parsed_json: PrimitiveJsonType) -> SerializableType:
        deserializer = self._create_deserializer()
        return deserializer.deserialize(parsed_json)

    def _create_deserializer(self) -> JsonObjectDeserializer:
        """
        Creates a deserializer that is to be used by this decoder.
        :return: the deserializer
        """
        if self._deserializer_cache is None:
            deserializer_cls = type(
                "%sInternalDeserializer" % type(self),
                (JsonObjectDeserializer,),
                {
                    "_JSON_ENCODER_ARGS": self._args,
                    "_JSON_ENCODER_KWARGS": self._kwargs
                }
            )
            self._deserializer_cache = deserializer_cls(self._get_property_mappings(), self._get_deserializable_cls())
        return self._deserializer_cache



from abc import ABCMeta


class BaseModel(metaclass=ABCMeta):
    """
    TODO
    """
    def __str__(self) -> str:
        string_builder = []
        for property, value in vars(self).items():
            string_builder.append("%s: %s" % (property, str(value)))
        string_builder = sorted(string_builder)
        return "{ %s }" % ', '.join(string_builder)

    def __repr__(self) -> str:
        return str(self)

    def __eq__(self, other) -> bool:
        if type(other) != type(self):
            return False
        for property, value in vars(self).items():
            other_value = getattr(other, property, object())
            if other_value != value:
                return False
        return True

    def __hash__(self):
        return hash(str(self))


class SimpleModel(BaseModel):
    """
    Example of a simple model.
    """
    def __init__(self, constructor_b=None):
        self.a = None
        self.b = constructor_b


class ComplexModel(SimpleModel):
    """
    Example of a complex model.
    """
    def __init__(self, constructor_b):
        super().__init__(constructor_b)
        self.c = None
        self.d = [SimpleModel(constructor_b + i) for i in range(3)]
        self.e = [4.1, 5.2, 6.3]
        self.f = {"1": 2, "3": 4}
        self.g = "test"
        self.h = 123
        self.i = {3, 1, 4, 1, 5, 9}

        for i in range(len(self.d)):
            self.d[i].a = i



from typing import Any

from hgijson.serialization import Serializer, Deserializer
from hgijson.custom_types import PrimitiveJsonType


class PrimitiveSerializer(Serializer):
    """
    Serializer for primitive values - just returns them.
    """
    def __init__(self, *args, **kwargs):
        super().__init__([])

    def serialize(self, serializable: Any) -> Any:
        return serializable

    def _create_serializer_of_type(self, serializer_type: type) -> None:
        """
        Unused - implemented to satisfy the interface only.
        """

    def _create_serialized_container(self) -> None:
        """
        Unused - implemented to satisfy the interface only.
        """


class PrimitiveDeserializer(Deserializer):
    """
    Deserializer for primitive values - just returns them.
    """
    def __init__(self, *args, **kwargs):
        super().__init__([], object)

    def deserialize(self, object_property_value_dict: PrimitiveJsonType) -> PrimitiveJsonType:
        return object_property_value_dict

    def _create_deserializer_of_type(self, deserializer_type: type) -> None:
        """
        Unused - implemented to satisfy the interface only.
        """



from setuptools import setup, find_packages

try:
    from pypandoc import convert
    def read_markdown(file: str) -> str:
        return convert(file, "rst")
except ImportError:
    def read_markdown(file: str) -> str:
        return open(file, "r").read()

setup(
    name="hgijson",
    version="3.0.1",
    author="Colin Nolan",
    author_email="colin.nolan@sanger.ac.uk",
    packages=find_packages(exclude=["tests"]),
    install_requires = open("requirements.txt", "r").readlines(),
    url="https://github.com/wtsi-hgi/python-json",
    license="MIT",
    description="Python 3 library for easily JSON encoding/decoding complex class-based Python models, using an "
                "arbitrarily complex (but easy to write!) mapping schema.",
    long_description=read_markdown("README.md"),
    keywords=["json", "serialization"],
    test_suite="hgijson.tests",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License"
    ]
)



import unittest

from hgijson.serializers import PrimitiveSerializer, PrimitiveDeserializer


class TestPrimitiveSerializer(unittest.TestCase):
    """
    Tests for `PrimitiveSerializer`.
    """
    def test_serialize(self):
        serializer = PrimitiveSerializer()
        self.assertDictEqual(serializer.serialize({"a": 1}), {"a": 1})
        self.assertEqual(serializer.serialize([1, 2]), [1, 2])
        self.assertEqual(serializer.serialize((1, 2)), (1, 2))
        self.assertEqual(serializer.serialize("a"), "a")
        self.assertEqual(serializer.serialize(5), 5)
        self.assertEqual(serializer.serialize(5.5), 5.5)
        self.assertEqual(serializer.serialize(True), True)
        self.assertEqual(serializer.serialize(False), False)
        self.assertEqual(serializer.serialize(None), None)


class TestPrimitiveDeserializer(unittest.TestCase):
    """
    Tests for `PrimitiveDeserializer`.
    """
    def test_deserialize(self):
        deserializer = PrimitiveDeserializer()
        self.assertDictEqual(deserializer.deserialize({"a": 1}), {"a": 1})
        self.assertEqual(deserializer.deserialize([1, 2]), [1, 2])
        self.assertEqual(deserializer.deserialize((1, 2)), (1, 2))
        self.assertEqual(deserializer.deserialize("a"), "a")
        self.assertEqual(deserializer.deserialize(5), 5)
        self.assertEqual(deserializer.deserialize(5.5), 5.5)
        self.assertEqual(deserializer.deserialize(True), True)
        self.assertEqual(deserializer.deserialize(False), False)
        self.assertEqual(deserializer.deserialize(None), None)


if __name__ == "__main__":
    unittest.main()
